/**
 * @file tinymt64-host.c
 *
 * @brief Sample Program for CUDA 2.3
 *
 * tinymt64
 * This program generates 32-bit unsigned integers.
 * The period of generated integers is 2<sup>127</sup>-1.
 *
 * This also generates single precision floating point numbers
 * uniformly distributed in the range [1, 2). (float r; 1.0 <= r < 2.0)
 *
 * @author Mutsuo Saito (Hiroshima University)
 * @author Makoto Matsumoto (The University of Tokyo)
 *
 * Copyright (C) 2010 Mutsuo Saito, Makoto Matsumoto,
 * Hiroshima University and The University of Tokyo.
 * All rights reserved.
 *
 * The new BSD License is applied to this software, see LICENSE.txt
 */
#include <stdio.h>
#include <stdint.h>
#include <inttypes.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include "tinymt64_host.h"

static int read_line(uint32_t *mat1, uint32_t *mat2, uint64_t *tmat, FILE *fp);

static int read_line(uint32_t *mat1, uint32_t *mat2, uint64_t *tmat, FILE *fp)
{
#define BUFF_SIZE 500
    char buff[BUFF_SIZE];
    char * p;
    uint32_t num;
    uint64_t number;
    errno = 0;
    for (;;) {
	if (feof(fp) || ferror(fp)) {
	    return -1;
	}
	fgets(buff, BUFF_SIZE, fp);
	if (errno) {
	    return errno;
	}
	if (buff[0] != '#') {
	    break;
	}
    }
    p = buff;
    for (int i = 0; i < 3; i++) {
	p = strchr(p, ',');
	if (p == NULL) {
	    return -1;
	}
	p++;
    }
    num = strtoul(p, &p, 16);
    if (errno) {
	return errno;
    }
    *mat1 = num;
    p++;
    num = strtoul(p, &p, 16);
    if (errno) {
	return errno;
    }
    *mat2 = num;
    p++;
    number = strtoull(p, &p, 16);
    if (errno) {
	return errno;
    }
    *tmat = number;
    return 0;
}

/**
 * This function sets constants in device memory.
 * @param[in] params input, tinymt64 parameters.
 */
int tinymt64_set_params(const char * filename,
			uint64_t * params,
			int num_param)
{
    FILE *ifp;
    int rc;
    uint32_t mat1 = 0;
    uint32_t mat2 = 0;
    uint64_t tmat = 0;
    int i;
    ifp = fopen(filename, "r");
    if (ifp == NULL) {
	return -1;
    }
    for (i = 0; i < num_param; i++) {
	rc = read_line(&mat1, &mat2, &tmat, ifp);
	if (rc != 0) {
	    return -2;
	}
	params[i * 3 + 0] = mat1;
	params[i * 3 + 1] = mat2;
	params[i * 3 + 2] = tmat;
    }
    fclose(ifp);
    return 0;
}
